import meep as mp
import argparse
import math
import cmath
import numpy as np
import matplotlib.pyplot as plt

# ***********************************
# Plane Wave (PW) amplitude function for oblique incidence!
# 
#  amplitude_PW returns the amplitude exp(ik(x+x0)) at a
#  given point x.  (x0 is the shift relative to the center of
#  the source (defined by MEEP)
#
#  amplitude_PW is a function of the wave vector k
#  and the displacement x0. It returns a function of x.
#
def amplitude_PW(k_vec, x0_vec):
    def _amplitude_PW(x_vec):
        return cmath.exp(1j*k_vec.dot(x_vec + x0_vec))
    return _amplitude_PW

def main(args):

    print("*** tanh_taper.py")

    resolution = args.res   # pixels/μm

    # ***********************************
    # tolerance
    # ***********************************
    tol = args.tol

    print("*** ---------------------------------- ***")
    print("*** Tolerance                          ***")
    print("*** ---------------------------------- ***")
    print("*** tol = {}".format(tol))

    # ***********************************
    # default material
    # ***********************************
    default_index = 1
    default_eps = default_index**2
    default_medium = mp.Medium(epsilon=default_eps)


    # ***********************************
    # Perfectly Matched Layers
    # ***********************************
    dpml_x = args.d_pmlX # length of PML in x direction
    dpml_y = args.d_pmlY # length of PML in y direction
    
    pml_layers = [mp.PML(dpml_x,direction=mp.X),
                  mp.PML(dpml_y,direction=mp.Y)]

    # ***********************************
    # Taper parameters
    # ***********************************
    <PLACEHOLDER>

    print("*** ---------------------------------- ***");
    print("*** TANH Taper                              ***")
    print("*** ---------------------------------- ***")
    print("*** tp_w1 = {}".format(tp_w1));
    print("*** tp_w2 = {}".format(tp_w2));
    print("*** ---------------------------------- ***");
    print("*** Waveguide                          ***")
    print("*** ---------------------------------- ***")
    print("*** wg_L = {}".format(wg_L));
    print("*** ---------------------------------- ***");
    print("*** Waveguide                          ***")
    print("*** ---------------------------------- ***")
    print("*** wg_eps = {}".format(wg_eps));
    print("*** ---------------------------------- ***");

    # ***********************************
    # Cell Size
    # ***********************************
    dair = args.d_Air    # length of air region

    <PLACEHOLDER>

    cell = mp.Vector3(sx,sy,sz)

    print("*** ---------------------------------- ***")
    print("*** Cell                               ***")
    print("*** ---------------------------------- ***")
    print("*** x = {}".format(sx))
    print("*** y = {}".format(sy))

    # ***********************************
    # Range of wavelengths/frequencies
    # ***********************************
    fcen = args.fcen        # center frequency
    lcen = 1/fcen           # center wavelength

    df = args.df
    lmin = 1/(fcen+0.5*df)  # min wavelength
    lmax = 1/(fcen-0.5*df)  # max wavelength
    lmid = 0.5*(lmin+lmax)  # mid wavelength

    fmin = fcen-0.5*df      # min frequency
    fmax = fcen+0.5*df      # max frequency
    fmid = 0.5*(fmin+fmax)  # min frequency

    nfreq = args.nfreq      # number of frequency bins

    print("*** ---------------------------------- ***")
    print("*** Source(s)                          ***")
    print("*** ---------------------------------- ***")
    print("*** lcen = {}".format(lcen))
    print("*** lmin = {}".format(lmin))
    print("*** lmid = {}".format(lmid))
    print("*** lmax = {}".format(lmax))
    print("*** dl   = {}".format(lmax-lmin))
    print("*** ---------------------------------- ***")
    print("*** fcen = {}".format(fcen))
    print("*** fmin = {}".format(fmin))
    print("*** fmid = {}".format(fmid))
    print("*** fmax = {}".format(fmax))
    print("*** df   = {}".format(df))
    print("*** ---------------------------------- ***")

    # ***********************************
    # Angle of incidence
    # ***********************************
    # rotation angle (in degrees) of source: 
    #  CCW around Y axis, 0 degrees along +Z axis
    theta_rad = math.radians(args.theta)

    # ***********************************
    # k_dir defines the direction
    #  the length has no meaning so far!
    #  plane of incidence is xz
    # ***********************************
    kx = math.cos(theta_rad)
    ky = math.sin(theta_rad)
    kz = 0
    k_dir = mp.Vector3(kx,ky,kz)

    # ***********************************
    # scale length of k
    #  2*pi*n*f = 2*pi*n/lambda = n*k0
    #  where k0 is the vacuum wave vector
    # ***********************************
    k_vec = k_dir.unit().scale(2*math.pi*default_index*fcen)

    # ***********************************
    # USE SYMMETRY, IF THETA IS ZERO ...
    # ***********************************
    symmetries = [mp.Mirror(mp.Y)]
    eig_parity = mp.ODD_Z+mp.EVEN_Y
    #eig_parity = mp.EVEN_Z
    #if args.theta == 0:
    #    k_vec = mp.Vector3(0,0,0)
    #    symmetries = [mp.Mirror(mp.Y)]
    #    eig_parity += mp.EVEN_Y


    # ***********************************
    # components of E-field and H-field
    # ***********************************
    comp_E = mp.Ez
    comp_H = mp.Hy

    # ***********************************
    # Sources
    # ***********************************

    print("*** Source is GaussBeam EigenModeSource")

    src_xpos = -0.5*sx+dpml_x+0.2*wg_L
    src_ypos = 0
    src_zpos = 0
    src_pos_vec = mp.Vector3(src_xpos,src_ypos,src_zpos)

    src_xsiz = 0
    src_ysiz = sy-2*dpml_y
    src_zsiz = 0
    src_siz_vec = mp.Vector3(src_xsiz,src_ysiz,src_zsiz)

    print("*** Source located at: {},{},{}".format(src_xpos,src_ypos,src_zpos))
    print("*** Source size is: {},{},{}".format(src_xsiz,src_ysiz,src_zsiz))

    sources = [ mp.EigenModeSource(mp.GaussianSource(frequency=fcen,fwidth=df), 
                                   center=src_pos_vec,
                                   size=src_siz_vec,
                                   component=comp_E,
                                   #amp_func=amplitude_PW(k_vec,src_pos_vec),
                                   eig_match_freq=True,
                                   eig_parity=eig_parity) ]

    # ***********************************
    # waveguide
    # ***********************************
    vertices = [ <PLACEHOLDER> ]

    # ***********************************
    # Simulation class for normalization run
    # ***********************************
    sim = mp.Simulation(resolution=resolution,
                        cell_size=cell,
                        boundary_layers=pml_layers,
                        geometry=[mp.Prism(vertices,height=mp.inf,material=Si)],
                        sources=sources,
                        #k_point=k_vec,
                        #default_material=default_medium,
                        symmetries = symmetries)

    # ***********************************
    # Monitoring point and FluxRegion
    # ***********************************
    in_mon_pt_pos_x = <PLACEHOLDER>
    in_mon_pt_pos_y = <PLACEHOLDER>
    in_mon_pt_pos_z = <PLACEHOLDER>
    in_mon_pt_pos = mp.Vector3(in_mon_pt_pos_x,in_mon_pt_pos_y)

    in_mon_pt_siz_x = <PLACEHOLDER>
    in_mon_pt_siz_y = <PLACEHOLDER>
    in_mon_pt_siz_z = <PLACEHOLDER>
    in_mon_pt_siz = mp.Vector3(in_mon_pt_siz_x,in_mon_pt_siz_y)

    n_flux = sim.add_flux(fcen,df,nfreq,mp.FluxRegion(center=in_mon_pt_pos,size=in_mon_pt_siz))

    # ***********************
    # NORMALIZATION RUN
    # ***********************
    sim.use_output_directory(args.dir)

    #if use_cw_solver:
    #    sim.init_sim()
    #    sim.solve_cw(tol, max_iters, L)
    #else:
    sim.run(mp.at_beginning(mp.output_epsilon),
            until_after_sources=mp.stop_when_fields_decayed(50,comp_E,in_mon_pt_pos,tol))
            #mp.at_every(10,mp.output_png(comp_E,"-C $EPS -Zc dkbluered")),            
            
    n_eigen_coeff = sim.get_eigenmode_coefficients(n_flux,[1],eig_parity=eig_parity)
    n_eigen_coeff_alpha = n_eigen_coeff.alpha
    
    # ***********************
    # Flux from normalization run
    # ***********************
    nflux = mp.get_fluxes(n_flux)
    nflux_freqs = mp.get_flux_freqs(n_flux)
    nflux_data = sim.get_flux_data(n_flux)
    
    # ***********************
    # ***********************
    # ***********************
    # ***********************

    # ***********************
    # TAPER RUN
    # ***********************
    sim.reset_meep()

    # ***********************************
    # taper geometry
    # ***********************************

    # linear taper
    vertices = [ <PLACEHOLDER> ]

    # ***********************************
    # Simulation class
    # ***********************************
    sim = mp.Simulation(resolution=resolution,
                        cell_size=cell,
                        boundary_layers=pml_layers,
                        geometry=[mp.Prism(vertices,height=mp.inf,material=Si)],
                        sources=sources,
                        #k_point=k_vec,
                        #default_material=default_medium,
                        symmetries = symmetries)

    # ***********************************
    # Load Minus Data for FluxRegion IN
    # ***********************************
    in_flux = sim.add_flux(fcen,df,nfreq,mp.FluxRegion(center=in_mon_pt_pos,size=in_mon_pt_siz))
    sim.load_minus_flux_data(in_flux,nflux_data)
    
    # ***********************************
    # Monitoring point and FluxRegion OUT
    # ***********************************
    out_mon_pt_pos_x = <PLACEHOLDER>
    out_mon_pt_pos_y = <PLACEHOLDER>
    out_mon_pt_pos_z = <PLACEHOLDER>
    out_mon_pt_pos = mp.Vector3(out_mon_pt_pos_x,out_mon_pt_pos_y)

    out_mon_pt_siz_x = <PLACEHOLDER>
    out_mon_pt_siz_y = <PLACEHOLDER>
    out_mon_pt_siz_z = <PLACEHOLDER>
    out_mon_pt_siz = mp.Vector3(out_mon_pt_siz_x,out_mon_pt_siz_y)

    out_flux = sim.add_flux(fcen,df,nfreq,mp.FluxRegion(center=out_mon_pt_pos,size=out_mon_pt_siz))

    # ***********************************
    # Subtract incident flux by loading 
    #  inverted flux from normalization run
    # ***********************************
    sim.use_output_directory(args.dir);

    #if use_cw_solver:
    #    sim.init_sim()
    #    sim.solve_cw(tol, max_iters, L)
    #else:
    sim.run(mp.at_beginning(mp.output_epsilon),
            mp.at_every(0.6,mp.output_png(comp_E,"-C $EPS -Zc dkbluered")),
            #mp.at_every(5,mp.output_efield_z),
            until_after_sources=mp.stop_when_fields_decayed(50,comp_E,in_mon_pt_pos,args.tol))

    # ***********************************
    # Fluxes and Frequencies at input side of taper
    # ***********************************
    in_eigen_coeff = sim.get_eigenmode_coefficients(in_flux,[1],eig_parity=eig_parity)
    in_eigen_coeff_alpha = in_eigen_coeff.alpha

    iflux = mp.get_fluxes(in_flux)
    iflux_freqs = mp.get_flux_freqs(in_flux)
    iflux_data = sim.get_flux_data(in_flux)

    # ***********************************
    # Fluxes and Frequencies at output side of taper
    # ***********************************
    out_eigen_coeff = sim.get_eigenmode_coefficients(out_flux,[1],eig_parity=eig_parity)
    out_eigen_coeffs_alpha = out_eigen_coeff.alpha

    oflux = mp.get_fluxes(out_flux)
    oflux_freqs = mp.get_flux_freqs(out_flux)
    oflux_data = sim.get_flux_data(out_flux)

    print("refl:, {}, {:.8f}, {:.8f}".format(tp_L,
	abs(out_eigen_coeffs_alpha[0,0,1])**2/abs(n_eigen_coeff_alpha[0,0,0])**2,
	oflux[0]/nflux[0]))

    # ***********************************
    # Do some postprocessing
    #  and print results to stdOut
    # ***********************************
    eps_data = sim.get_array(center=mp.Vector3(), size=cell_plt, component=mp.Dielectric)
    plt.figure(dpi=100)
    plt.imshow(eps_data.transpose(), interpolation='spline36', cmap='RdBu')
    plt.axis('on')
    #plt.show()

    fname = args.dir+"/eps.png"
    plt.savefig(fname,dpi=100)

    # ***********************************
    # Ez
    # ***********************************
    ez_data = sim.get_array(center=mp.Vector3(0,0,0), size=cell_plt, component=mp.Ez.real)
    plt.figure(dpi=100)
    plt.imshow(eps_data.transpose(), interpolation='spline36', cmap='binary')
    plt.imshow(ez_data.real.transpose(), interpolation='spline36', cmap='RdBu', alpha=0.9)
    plt.axis('off')
    #plt.show()

    fname = args.dir+"/ez.png"
    plt.savefig(fname,dpi=100)

    # DO THE SAME FOR EY and EX

    # ***********************************
    # Ey
    # ***********************************
    <PLACEHOLDER>

    fname = args.dir+"/ey.png"
    plt.savefig(fname,dpi=100)

    # ***********************************
    # Ex
    # ***********************************
    <PLACEHOLDER>

    fname = args.dir+"/ex.png"
    plt.savefig(fname,dpi=100)

    print("*** ---------------------------------- ***")
    print("i, iflux_freqs[i], iflux[i]")
    for i in range(nfreq):
        print("{}, {}, {}".format(i,iflux_freqs[i],iflux[i]))

    # DO THE SAME FOR THE OUTPUT FLUX

    print("*** ---------------------------------- ***")
    print("i, oflux_freqs[i], oflux[i]")
    <PLACEHOLDER>


    # ENERGY FLUX ANALYSIS

    In_plt = np.zeros(nfreq) 
    Out_plt = np.zeros(nfreq) 
    Loss_plt = np.zeros(nfreq) 
    Err_plt = np.zeros(nfreq) 
    Sum_plt = np.zeros(nfreq) 

    th_i_plt = np.zeros(nfreq) # theta
    th_o_plt = np.zeros(nfreq) # theta

    la_i_plt = np.zeros(nfreq) # lambda 
    la_o_plt = np.zeros(nfreq) # lambda 

    loss = np.zeros(nfreq)

    print("*** ---------------------------------- ***")
    print("*** Flux at input side od taper ")
    print("i: i, freq(i), lambda(i), theta(i), iflux(i)/iflux(i) ")
    for i in range(nfreq):
        la_i_plt[i] = 1/iflux_freqs[i] if iflux_freqs[i] != 0 else 0
        th_i_plt[i] = math.degrees(math.asin(k_vec.y/iflux_freqs[i])) if ( (k_vec.y<=iflux_freqs[i]) & (iflux_freqs[i]!=0) ) else 0 # remove over-evanescent waves
        In_plt[i] = iflux[i]/nflux[i] if nflux[i] != 0 else 0
        print("i:, {}, {}, {}, {}, {}".format(i,iflux_freqs[i],la_i_plt[i],th_i_plt[i],In_plt[i]))

    # DO THE SAME FOR THE OUTPUT SIDE OF THE TAPER

    print("*** ---------------------------------- ***")
    print("*** Flux at output side of taper ")
    print("t: i, freq(i), lambda(i), theta(i), oflux(i)/iflux(i) ")
    <PLACEHOLDER>

    # COMPUTE THE LOSS AND PRINT TO OUTPUT

    print("*** ---------------------------------- ***")
    print("*** Loss ")
    print("a: i, freq(i), lambda(i), theta(i), loss(i) ")
    for i in range(nfreq):
        loss[i] =  <PLACEHOLDER>
        Loss_plt[i] = <PLACEHOLDER> if nflux[i]!=0 else 0
        print("a:, {}, {}, {}, {}, {}".format(i,oflux_freqs[i],la_o_plt[i],th_o_plt[i],Loss_plt[i]))

    # PRINT THE FLUX SUMMARY

    print("*** ---------------------------------- ***")
    print("*** Flux Summary ***") 
    print("s:,  i, iflux[i], oflux[i], loss[i], Sum[i], Err[i]")
    for i in range(nfreq):
        Sum_plt[i] = ( <PLACEHOLDER> ) / nflux[i]
        Err_plt[i] = <PLACEHOLDER>
        print("s:, {}, {}, {}, {}, {}, {}".format(i,In_plt[i],Out_plt[i],Loss_plt[i],Sum_plt[i],Err_plt[i]))

    print("*** ---------------------------------- ***")
    print("*** Error Summary ***") 
    print("e:    i, Err[i]")
    for i in range(nfreq):
        print("e:, {}, {}".format(i,Err_plt[i]))

    # ***********************************
    # Flux Summary
    # ***********************************
    plt.figure(dpi=100)
    plt.title('Flux Summary')
    plt.xlabel('Lambda [um]')
    plt.ylabel('Flux [W/m/m]')
    plt.plot(la_i_plt,Sum_plt,'k^',
             la_i_plt,Err_plt,'r^',
             la_i_plt,In_plt,'r-',
             la_o_plt,Out_plt,'g-',
             la_o_plt,Loss_plt,'b--'
             )
    #plt.text(lmid,0.5,r'$...$')
    plt.axis([1/np.max(iflux_freqs),1/np.min(iflux_freqs),0,1.05])
    plt.axis('on')
    plt.grid(True)
    #plt.show()

    fname = args.dir+"/efluxSummary.png"
    plt.savefig(fname,dpi=100)

    # ***********************************
    # Empty flux
    # ***********************************
    plt.figure(dpi=100)
    plt.title('Reflected Flux')
    plt.xlabel('Lambda [um]')
    plt.ylabel('Reflectance [W/m/m]')
    plt.plot(la_i_plt,In_plt,'ko')
    #plt.text(lmid,0.5,r'$...$')
    plt.axis([lmin,lmax,0,1.1])
    plt.axis('on')
    plt.grid(True)
    #plt.show()

    fname = args.dir+"/efluxI.png"
    plt.savefig(fname,dpi=100)

    # ***********************************
    # Transmittance
    # ***********************************
    plt.figure(dpi=100)
    plt.title('Transmitted Flux')
    plt.xlabel('Lambda [um]')
    plt.ylabel('Transmittance [-]')
    plt.plot(la_o_plt,Out_plt,'k-')
    #plt.text(lmid,0.5,r'$...$')
    plt.axis([1/np.max(oflux_freqs),1/np.min(oflux_freqs),0,1.1])
    plt.axis('on')
    plt.grid(True)
    #plt.show()

    fname = args.dir+"/efluxT.png"
    plt.savefig(fname,dpi=100)

    # ***********************************
    # Loss
    # ***********************************
    plt.figure(dpi=100)
    plt.title('Taper Loss')
    plt.xlabel('Lambda [um]')
    plt.ylabel('Loss [-]')
    plt.plot(la_o_plt,Loss_plt,'k-')
    #plt.text(lmid,0.5,r'$T(\lambda)\ over\ \lambda$')
    plt.axis([1/np.max(iflux_freqs),1/np.min(iflux_freqs),0,1.1])
    plt.axis('on')
    plt.grid(True)
    #plt.show()

    fname = args.dir+"/efluxA.png"
    plt.savefig(fname,dpi=100)

    # ***********************************
    # Error
    # ***********************************
    plt.figure(dpi=100)
    plt.title('Calculation Error')
    plt.xlabel('Lambda [um]')
    plt.ylabel('Error [-]')
    plt.plot(la_o_plt,Err_plt,'ro')
    #plt.text(lmid,0.5,r'$T(\lambda)\ over\ \lambda$')
    plt.axis([1/np.max(iflux_freqs),1/np.min(iflux_freqs),0,1.1])
    plt.axis('on')
    plt.grid(True)
    #plt.show()

    fname = args.dir+"/efluxErr.png"
    plt.savefig(fname,dpi=100)

    # ***********************************
    # Eigenmode Analysis
    # ***********************************
    mode_num = 1



    print("*** ---------------------------------- ***")
    print("Number of Modes:, {}".format(mode_num))
    In_eig_modes = sim.get_eigenmode_coefficients(in_flux, range(1,mode_num+1), eig_parity=eig_parity)
    In_eig_coeffs = In_eig_modes.alpha

    print("*** ---------------------------------- ***")
    print("eigModes(I):, i, angle, mode")

    Isum = 0
    for i in range(mode_num):
      In_kdom = In_eig_modes.kdom[i]
      In_mode = abs(In_eig_coeffs[i,0,1])**2/nflux[0] if nflux[0] != 0 else 0
      In_angle = np.sign(In_kdom.y)*math.acos(In_kdom.x/(default_index*fcen)) if In_kdom.x<=(default_index*fcen) else 0
      print("eigModes(I):, {}, {:.2f}, {:.8f}".format(i,math.degrees(In_angle),In_mode))
      Isum += In_mode
 

    print("*** ---------------------------------- ***")
    print("Number of Modes:, {}".format(mode_num))
    Out_eig_modes = sim.get_eigenmode_coefficients(out_flux, range(1,mode_num+1), eig_parity=eig_parity)
    Out_eig_coeffs = Out_eig_modes.alpha

    print("*** ---------------------------------- ***")
    print("eigModes(O):, i, angle, mode")

    Tsum = 0
    for i in range(mode_num):
      Out_kdom = Out_eig_modes.kdom[i]
      Out_mode = abs(Out_eig_coeffs[i,0,1])**2/nflux[0] if nflux[0] != 0 else 0
      Out_angle = np.sign(Out_kdom.y)*math.acos(Out_kdom.x/(default_index*fcen)) if Out_kdom.x<=(default_index*fcen) else 0
      print("eigModes(Tap):, {}, {:.2f}, {:.8f}".format(i,math.degrees(Out_angle),Out_mode))
      Tsum += Out_mode


    print("*** ---------------------------------- ***")
    print("modeFlux:, I={:.6f}, T={:.6f}, I-T={:.6f}, 1-(I-T)={}".format(Isum,Tsum,Isum+Tsum,1-(Isum+Tsum)))

    # ***********************************
    # Poynting Flux
    # ***********************************
    oflux_at_zero = oflux[0]/nflux[0]
    print("*** ---------------------------------- ***")
    print("poyntingFlux:, T={:.6f}".format(oflux_at_zero))

    print("*** -------------END------------------ ***")


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-res', type=int, default=61, help='resolution (default: 61 pixels/um)')
    parser.add_argument('-fcen', type=float, default=0.15, help='center frequency (default: 0.1499 1/um)')
    parser.add_argument('-df', type=float, default=0.2, help='spectral range (default: 0.5 1/um)')
    parser.add_argument('-theta', type=float, default=0, help='angle of incidence (default: 0 degrees)')
    parser.add_argument('-nfreq', type=int, default=20, help='number of frequencies (default: 5)')
    # DEFINE YOUR PARAMETERS HERE ... reference for example by args.res in this Python script
    <PLACEHOLDER>
    parser.add_argument('-dir', type=str, default='./', help='output directory (default ./)')
    parser.add_argument('-tol', type=float, default=1e-3, help='tolerance (default 1e-3)')
    args = parser.parse_args()
    main(args)
